import shutil
import torch


class AverageMeter(object):
    """Computes and stores the average and current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def accuracy(output, target, topk=(1,)):
    """Computes the precision@k for the specified values of k"""
    maxk = max(topk)
    batch_size = target.size(0)
    _, pred = output.topk(maxk, 1, largest=True, sorted=True)
    pred = pred.t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))

    res = []
    for k in topk:
        correct_k = correct[:k].contiguous().view(-1).float().sum(0, keepdim=True)
        res.append(correct_k.mul_(100.0 / batch_size))
    return res


def save_checkpoint(state, is_best, filename='alex_checkpoint.pth'):
    torch.save(state, filename)
    if is_best:
        shutil.copyfile(filename, 'alex_model_best.pth')


def adjust_learning_rate(optimizer, epoch, init_lr , start_epoch = 0):
    """Sets the learning rate to the initial LR decayed by 10 every 30 epochs"""
    if(start_epoch == 0):
        lr = init_lr * (0.1 ** (epoch // 30))
    else:
        lr = init_lr * (0.1 ** (start_epoch // 30))
        lr = lr * (0.8 ** ((epoch - start_epoch) // 50))
    for param_group in optimizer.param_groups:
        param_group['lr'] = lr
